# -*- coding: utf-8 -*-
"""
/*
 * Gravitational-orbitals.c
 * Copyright 2025 Malcolm J. Macleod (malcolm@codingthecosmos.com)
 *
 * This software is free to use, modify, and distribute. However, any derivative works
 * or redistributions must include appropriate credit to Malcolm Macleod and this permission notice shall
 * be included in all copies or substantial portions of the Software.
 */
 
"""
import numpy as np
import matplotlib.pyplot as plt

class Body:
    def __init__(self, x, y, vx, vy):
        self.position = np.array([x, y], dtype=np.float64)
        self.velocity = np.array([vx, vy], dtype=np.float64)

class ThreeBodySystem:
    def __init__(self, bodies, G=1.453):
        if len(bodies) != 3:
            raise ValueError("ThreeBodySystem requires exactly 3 Body objects")
        self.bodies = bodies
        self.G = G
        self.min_distance = 1e-10

    def calculate_forces(self):
        forces = [np.zeros(2) for _ in range(3)]
        for i in range(3):
            for j in range(i+1, 3):
                r_vec = self.bodies[j].position - self.bodies[i].position
                r_mag = max(np.linalg.norm(r_vec), self.min_distance)
                force_mag = self.G / r_mag**2
                force_dir = r_vec / r_mag
                forces[i] += force_mag * force_dir
                forces[j] -= force_mag * force_dir
        return forces

    def calculate_energy(self):
        kinetic = 0.5 * sum(np.sum(body.velocity**2) for body in self.bodies)
        potential = 0.0
        for i in range(3):
            for j in range(i+1, 3):
                r = max(np.linalg.norm(self.bodies[j].position - self.bodies[i].position),
                        self.min_distance)
                potential -= self.G / r
        return kinetic + potential

    def calculate_angular_momentum(self):
        return sum(body.position[0]*body.velocity[1] - body.position[1]*body.velocity[0]
               for body in self.bodies)

    def update_leapfrog(self, dt=1.0):
        forces = self.calculate_forces()
        for i, body in enumerate(self.bodies):
            body.velocity += 0.5 * forces[i] * dt
        for body in self.bodies:
            body.position += body.velocity * dt
        forces = self.calculate_forces()
        for i, body in enumerate(self.bodies):
            body.velocity += 0.5 * forces[i] * dt
        return self.get_positions()

    def get_positions(self):
        return np.concatenate([body.position for body in self.bodies])

def load_data_and_initialize_system(filename):
    data_initial = np.loadtxt(filename, delimiter=',', usecols=range(1,7), max_rows=2)
    data_orbital = np.loadtxt(filename, delimiter=',', usecols=range(1,7))
    
    vels = data_initial[1] - data_initial[0]
    bodies = [
        Body(*data_initial[0, 0:2], *vels[0:2]),
        Body(*data_initial[0, 2:4], *vels[2:4]),
        Body(*data_initial[0, 4:6], *vels[4:6])
    ]
    return ThreeBodySystem(bodies), data_orbital

def generate_newtonian_data(system, n_iterations):
    positions = np.zeros((n_iterations, 6))
    energy = np.zeros(n_iterations)
    angular_momentum = np.zeros(n_iterations)
    
    positions[0] = system.get_positions()
    energy[0] = system.calculate_energy()
    angular_momentum[0] = system.calculate_angular_momentum()
    
    for i in range(1, n_iterations):
        positions[i] = system.update_leapfrog(1.0)
        energy[i] = system.calculate_energy()
        angular_momentum[i] = system.calculate_angular_momentum()
    
    np.savetxt("newtonian_results.csv", 
               np.hstack([np.arange(1, n_iterations+1)[:, None], positions]),
               delimiter=',', header="time,x1,y1,x2,y2,x3,y3",
               comments='', fmt=['%d'] + ['%.8f']*6)
    
    return positions, energy, angular_momentum

def calculate_orbital_metrics(orbital_data, G=1.453):
    n_steps = len(orbital_data)
    energy = np.zeros(n_steps)
    angular_momentum = np.zeros(n_steps)
    
    for i in range(n_steps):
        pos = orbital_data[i]
        x1, y1, x2, y2, x3, y3 = pos
        
        if i < n_steps - 1:
            vel = orbital_data[i+1] - pos
        else:
            vel = pos - orbital_data[i-1]
        vx1, vy1, vx2, vy2, vx3, vy3 = vel
        
        kinetic = 0.5 * (vx1**2 + vy1**2 + vx2**2 + vy2**2 + vx3**2 + vy3**2)
        r12 = np.hypot(x2-x1, y2-y1)
        r13 = np.hypot(x3-x1, y3-y1)
        r23 = np.hypot(x3-x2, y3-y2)
        potential = -G * (1/r12 + 1/r13 + 1/r23)
        energy[i] = kinetic + potential
        
        am = (x1*vy1 - y1*vx1) + (x2*vy2 - y2*vx2) + (x3*vy3 - y3*vx3)
        angular_momentum[i] = am
    
    return energy, angular_momentum

def plot_m1_distances(orbital_data, newton_data, indices):
    plt.figure(figsize=(12, 6))
    
    # Calculate distances
    orb_m1m2 = np.hypot(orbital_data[:,2] - orbital_data[:,0],
                        orbital_data[:,3] - orbital_data[:,1])
    orb_m1m3 = np.hypot(orbital_data[:,4] - orbital_data[:,0],
                        orbital_data[:,5] - orbital_data[:,1])
    new_m1m2 = np.hypot(newton_data[:,2] - newton_data[:,0],
                        newton_data[:,3] - newton_data[:,1])
    new_m1m3 = np.hypot(newton_data[:,4] - newton_data[:,0],
                        newton_data[:,5] - newton_data[:,1])
    
    #plt.plot(indices+1, orb_m1m2[indices], 'b-', label='Orbital m1-m2')
    #plt.plot(indices+1, orb_m1m3[indices], 'g-', label='Orbital m1-m3')
    plt.plot(indices+1, new_m1m2[indices], 'r--', label='Newtonian m1-m2')
    plt.plot(indices+1, new_m1m3[indices], 'm--', label='Newtonian m1-m3')
    
    plt.title('Distance from m1 to Other Bodies')
    plt.xlabel('Time (Iterations)')
    plt.ylabel('Distance')
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("m1_distances.png")

#def plot_m1_distances(orbital_data, newton_data, indices):
    plt.figure(figsize=(12, 6))
    
    # Calculate distances
    orb_m1m2 = np.hypot(orbital_data[:,2] - orbital_data[:,0],
                        orbital_data[:,3] - orbital_data[:,1])
    orb_m1m3 = np.hypot(orbital_data[:,4] - orbital_data[:,0],
                        orbital_data[:,5] - orbital_data[:,1])
    new_m1m2 = np.hypot(newton_data[:,2] - newton_data[:,0],
                        newton_data[:,3] - newton_data[:,1])
    new_m1m3 = np.hypot(newton_data[:,4] - newton_data[:,0],
                        newton_data[:,5] - newton_data[:,1])
    
    plt.plot(indices+1, orb_m1m2[indices], 'b-', label='Orbital m1-m2')
    plt.plot(indices+1, orb_m1m3[indices], 'g-', label='Orbital m1-m3')
    #plt.plot(indices+1, new_m1m2[indices], 'r--', label='Newtonian m1-m2')
    #plt.plot(indices+1, new_m1m3[indices], 'm--', label='Newtonian m1-m3')
    
    plt.title('Distance from m1 to Other Bodies')
    plt.xlabel('Time (Iterations)')
    plt.ylabel('Distance')
    plt.legend()
    plt.grid(True)
    plt.tight_layout()
    plt.savefig("m1_distances.png")
    
def plot_trajectory_comparison(orbital_data, newton_data, indices):
    plt.figure(figsize=(12, 12))
    
    plt.plot(orbital_data[indices, 0], orbital_data[indices, 1], 'b-', label='Orbital Body 1', alpha=0.7)
    plt.plot(orbital_data[indices, 2], orbital_data[indices, 3], 'g-', label='Orbital Body 2', alpha=0.7)
    #plt.plot(orbital_data[indices, 4], orbital_data[indices, 5], 'k-', label='Orbital Body 3', alpha=0.7)
    
    plt.plot(newton_data[indices, 0], newton_data[indices, 1], 'r--', label='Newtonian Body 1')
    plt.plot(newton_data[indices, 2], newton_data[indices, 3], 'm--', label='Newtonian Body 2')
    #plt.plot(newton_data[indices, 4], newton_data[indices, 5], 'c--', label='Newtonian Body 3')
    
    plt.xlabel('X Position')
    plt.ylabel('Y Position')
    plt.title('Trajectory Comparison')
    plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left')
    plt.grid(True)
    plt.tight_layout()
    plt.savefig('trajectory_comparison.png', bbox_inches='tight')

def main():
    TARGET_ITERATION = 262145
    FILENAME = "data-3b-extended.txt"
    
    system, orbital_data = load_data_and_initialize_system(FILENAME)
    newton_positions, newton_energy, newton_am = generate_newtonian_data(system, TARGET_ITERATION)
    
    sample_rate = max(1, TARGET_ITERATION // 1000)
    indices = np.arange(0, TARGET_ITERATION, sample_rate)
    
    # Generate plots
    plot_trajectory_comparison(orbital_data, newton_positions, indices)
    plot_m1_distances(orbital_data, newton_positions, indices)
    
    print("Analysis complete. Results saved as:")
    print("- trajectory_comparison.png")
    print("- m1_distances.png")
    print("- newtonian_results.csv")

if __name__ == "__main__":
    main()